/*#io
Blowfish ioDoc(
			docCopyright("Steve Dekorte", 2004)
			docLicense("BSD revised")
			docObject("Blowfish")
			docDescription("""The Blowfish object can be used to do encryption and decryption using the blowfish algorithm.
Example use;
<pre>
bf = Blowfish clone
bf key = "secret"
bf data = "this is a message" asBuffer
bf encrypt
// bf data now contains the encrypted message
</pre>
""")
			)
*/

#include "IoBlowfish.h"
#include "IoState.h"
#include "IoNumber.h"
#include "blowfish.h"
#include "IoSeq.h"

#define DATA(self) ((blowfish_ctx *)(self->data))

IoTag *IoBlowfish_tag(void *state)
{
	IoTag *tag = IoTag_newWithName_("Blowfish");
	tag->state = state;
	tag->freeFunc = (TagFreeFunc *)IoBlowfish_free;
	tag->cloneFunc = (TagCloneFunc *)IoBlowfish_rawClone;
	return tag;
}

IoBlowfish *IoBlowfish_proto(void *state)
{
	IoBlowfish *self = IoObject_new(state);
	self->tag = IoBlowfish_tag(state);
	
	self->data = calloc(1, sizeof(blowfish_ctx));
	
	IoState_registerProtoWithFunc_(state, self, IoBlowfish_proto);
	
	{
		IoMethodTable methodTable[] = {    
		{"encrypt", IoBlowfish_encrypt},
		{"decrypt", IoBlowfish_decrypt},
		{0x0, 0x0},
		};
		IoObject_addMethodTable_(self, methodTable);
	}
	
	IoObject_rawDoString_label_(self, 
						   IOSYMBOL("setKey := method(k, self key := k asBuffer)"), 
						   IOSYMBOL("[Blowfish]"));
	
	IoObject_rawDoString_label_(self, 
						   IOSYMBOL("setData := method(d, self data := d asBuffer; self size := data size)"), 
						   IOSYMBOL("[Blowfish]"));
	
	return self;
}

IoBlowfish *IoBlowfish_rawClone(IoBlowfish *proto) 
{ 
	IoObject *self = IoObject_rawClonePrimitive(proto);
	self->data = calloc(1, sizeof(blowfish_ctx));
	return self; 
}

IoBlowfish *IoBlowfish_new(void *state)
{
	IoObject *proto = IoState_protoWithInitFunction_(state, IoBlowfish_proto);
	return IOCLONE(proto);
}

/* ----------------------------------------------------------- */

void IoBlowfish_free(IoBlowfish *self) 
{ 
	free(DATA(self)); 
}

/* ----------------------------------------------------------- */

void blowfishEncrypt(ByteArray *data, ByteArray *key);
void blowfishDecrypt(ByteArray *data, ByteArray *key);

IoObject *IoBlowfish_encrypt(IoBlowfish *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("encrypt", 
		   "Encrypts the buffer in the data slot using the key in the key slot.")
	*/
	
	IoSeq *key  = IoObject_getSlot_(self, IOSYMBOL("key"));
	IoSeq *data = IoObject_getSlot_(self, IOSYMBOL("data"));
	IOASSERT(ISMUTABLESEQ(key), "Blowfish requires a mutable Sequence data slot");
	blowfishEncrypt(IoSeq_rawByteArray(data), IoSeq_rawByteArray(key));
	return self;
}

IoObject *IoBlowfish_decrypt(IoBlowfish *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("decrypt", 
		   "Decrypts the buffer in the data slot using the key in the key slot.")
	*/
	
	IoSeq *key  = IoObject_getSlot_(self, IOSYMBOL("key"));
	IoSeq *data = IoObject_getSlot_(self, IOSYMBOL("data"));
	IOASSERT(ISMUTABLESEQ(key), "Blowfish requires a mutable Sequence data slot");
	blowfishDecrypt(IoSeq_rawByteArray(data), IoSeq_rawByteArray(key));
	return self;
}

void blowfishEncrypt(ByteArray *data, ByteArray *key)
{
	blowfish_ctx    context;
	unsigned long   lr[2];
	unsigned int    i;
	unsigned int    n = ByteArray_size(data);
	unsigned char   *p = ByteArray_bytes(data);
	unsigned int    runs = n / sizeof(lr);
	
	blowfish_init(&context, ByteArray_bytes(key), ByteArray_size(key));
	
	for (i = 0; i < runs; i++)
	{
		memcpy(lr, p, sizeof(lr));
		blowfish_encrypt(&context, &lr[0], &lr[1]);
		memcpy(p, lr, sizeof(lr));
		p += sizeof(lr);
	}
	
	if (n -= runs * sizeof(lr))
	{
		memset(lr, 0x0, sizeof(lr));
		memcpy(lr, p, n);
		blowfish_encrypt(&context, &lr[0], &lr[1]);
		
		{
			size_t pos = p - ByteArray_bytes(data);
			ByteArray_setSize_(data, ByteArray_size(data) + sizeof(lr) - n);
			p = ByteArray_bytes(data) + pos;
		}
		
		memcpy(p, lr, sizeof(lr));
	}
}

void blowfishDecrypt(ByteArray *data, ByteArray *key)
{
	blowfish_ctx    context;
	unsigned long   lr[2];
	unsigned int    i;
	unsigned int    n = ByteArray_size(data);
	unsigned char   *p = (void *)ByteArray_bytes(data);
	unsigned int    runs = n / sizeof(lr);
	
	blowfish_init(&context, (unsigned char *)ByteArray_bytes(key), ByteArray_size(key));
	
	for (i = 0; i < runs; i++)
	{
		memcpy(lr, p, sizeof(lr));
		blowfish_decrypt(&context, &lr[0], &lr[1]);
		memcpy(p, lr, sizeof(lr));
		p += sizeof(lr);
	}
	
	if (n -= runs * sizeof(lr))
	{
		memset(lr, 0x0, sizeof(lr));
		memcpy(lr, p, n);
		blowfish_decrypt(&context, &lr[0], &lr[1]);
		memcpy(p, lr, n);
	}
	
	/* still might need to chop off extra bytes, but need to know original size to do that */
}
