/*
 Regex ioDoc(
		   docCopyright("Steve Dekorte", 2005)
		   docLicense("BSD revised")
		   docObject("Regex")
		   docDescription("""The Regex addon adds support for Perl regular expressions 
 using the <a href=http://www.pcre.org/>PCRE</a> library. (PCRE has a BSD-friendly license.)
 Example use;
 <pre>
 r := Regex clone setPattern("aa*")
 matches := r setString("11aabb") allMatches // returns "aa"
 Regex rSubstitute ("wombats are cuddly", "(wom)(bat)", "\\2\\1!") // 
 returns "batwom!s are cuddly"
 </pre>
 
 <br>
 <i>
 Some people, when confronted with a problem, think
 "I know, I'll use regular expressions." <br>
 Now they have two problems.</i><br>
 - Jamie Zawinski
 
 credits: PCRE library by Philip Hazel; rSubstitute method by Jason 
 Grossman""")
		   */


/*
 docSlot("allMatches")
 docDescription("Returns a List of strings containing all matches of 
 receiver's pattern within its string.")
 */

/*
 docSlot("caselessOn")
 docDescription("Turns case insensitive searching on.")
 */

/*
 docSlot("caselessOff")
 docDescription("Turns case insensitive searching off.")
 */

/*
 docSlot("clone")
 docDescription("Returns a copy of the receiver.")
 */

/*
 docSlot("dollarEndOnlyOn")
 docDescription("When on, a dollar metacharacter in the pattern matches 
 only at the end of the subject string. Without this option, a dollar 
 also matches immediately before the final character if it is a newline 
 (but not before any other newlines). The PCRE_DOLLAR_ENDONLY option is 
 ignored if PCRE_MULTILINE is set. There is no equivalent to this option ")
 in Perl.
 */

/*
 docSlot("dollarEndOnlyOff")
 docDescription("Turns the dollarEndOnly search mode off.")
 */

/*
 docSlot("firstMatch")
 docDescription("Returns a String containing the first match or a List 
 containing the first match group or Nil if no match is found.")
 */

/*
 docSlot("firstMatchRange")
 docDescription("Returns a List containing two numbers that are the index 
 of the first match and its length. Nil is returned if no match is 
 found.")
 */

/*
 docSlot("greedyOff")
 docDescription("Turns the greedy search mode off.")
 */

/*
 docSlot("greedyOn")
 docDescription("This option inverts the greediness of the quantifiers so 
 that they are not greedy by default, but become greedy if followed by 
 ?. It is not compatible with Perl. It can also be set by a (?U) 
 option setting within the pattern.")
 */

/*
 docSlot("hasMatch")
 docDescription("Returns true if there are any matches or false otherwise.")
 */

/*
 docSlot("matchEmptyOff")
 docDescription("Turns matchEmpty searching off.")
 */

/*
 docSlot("matchEmptyOn")
 docDescription("An empty string is not considered to be a valid match if 
 this option is set. If there are alternatives in the pattern, they are 
 tried. If all the alternatives match the empty string, the entire match 
 fails.")
 */

/*
 docSlot("multilineOff")
 docDescription("Turns the multiline search mode off.")
 */

/*
 docSlot("multilineOn")
 docDescription("By default, PCRE treats the subject string as consisting 
 of a single "line" of characters (even if it actually contains several 
							newlines). The "start of line" metacharacter (^) matches only at the 
 start of the string, while the "end of line" metacharacter ($) matches 
 only at the end of the string, or before a terminating newline (unless 
													PCRE_DOLLAR_ENDONLY is set). This is the same as Perl. When 
 PCRE_MULTILINE it is set, the "start of line" and "end of line" 
 constructs match immediately following or immediately before any 
 newline in the subject string, respectively, as well as at the very 
 start and end. This is equivalent to Perl's /m option. If there are no 
 "\n" characters in a subject string, or no occurrences of ^ or $ in a 
 pattern, setting PCRE_MULTILINE has no effect.")
 */

/*
 docSlot("nextMatch")
 docDescription("Returns a String containing the next match or a List 
 containing the next match group or Nil if no match is found.")
 */

/*
 docSlot("nextMatchRange")
 docDescription("Returns a List containing two numbers that are the index 
 of the next match and its length. Nil is returned if no match is found.")
 */

/*
 docSlot("pattern")
 docDescription("Returns the current pattern string.")
 */

/*
 docSlot("rSubstitute (aSequence, aString, aString)")
 docDescription("replaces all occurrences of the second argument (a 
													 pattern) with the third argument (a replacement string), with the 
 substring "\\0" in the replacement string replaced by a string matching 
 the whole pattern, and substrings "\\1", "\\2" etc. in the replacement 
 string replaced by the substrings matching the first, second etc. 
 brocketed expressions in the pattern.")
 */

/*
 docSlot("setPattern(aString)")
 docDescription("Sets (and compiles) the pattern to be used for finding 
 matches. The default pattern is "*". Returns self.")
 */

/*
 docSlot("setString(aString)")
 docDescription("Sets the string to find matches within.")
 */

/*
 docSlot("string")
 docDescription("Returns the string to find matches within.")
 */

/*
 docSlot("type")
 docDescription("Returns Regex.")
 */

/*
 docSlot("UTFOff")
 docDescription("Turns the UTF8 search mode off.")
 */

/*
 docSlot("UTF8On")
 docDescription("This option causes PCRE to regard both the pattern and the 
 subject as strings of UTF-8 characters instead of just byte strings. 
 However, it is available only if PCRE has been built to include UTF-8 
 support. If not, the use of this option provokes an error. Support for 
 UTF-8 is new, experimental, and incomplete. Details of exactly what it 
 entails are given below.")
 */

/*
 docSlot("version")
 docDescription("Returns a String with the version info of the underlying 
 library.")
 */



#include "IoRegex.h"
#include "IoState.h"
#include "IoNumber.h"
#include "IoList.h"
#include <stdlib.h>

#define DATA(self) ((IoRegexData *)self->data)

IoTag *IoRegex_tag(void *state)
{
	IoTag *tag = IoTag_newWithName_("Regex");
	tag->state = state;
	tag->freeFunc = (TagFreeFunc *)IoRegex_free;
	tag->cloneFunc = (TagCloneFunc *)IoRegex_rawClone;
	tag->markFunc = (TagMarkFunc *)IoRegex_mark;
	return tag;
}

IoRegex *IoRegex_proto(void *state)
{
	IoObject *self = IoObject_new(state);
	self->tag = IoRegex_tag(state);
	
	self->data = calloc(1, sizeof(IoRegexData));
	
	DATA(self)->pattern = IOSYMBOL("^([^']+)$");
	DATA(self)->string = IOSYMBOL("");
	DATA(self)->compileOptions = PCRE_MULTILINE;
	
	IoState_registerProtoWithFunc_(state, self, IoRegex_proto);
	
	{
		IoMethodTable methodTable[] = {
		{"setPattern", IoRegex_setPattern},
		{"pattern", IoRegex_pattern},
		{"setString", IoRegex_setString},
		{"string", IoRegex_string},
			
		{"firstMatchRange", IoRegex_firstMatchRange},
		{"nextMatchRange", IoRegex_nextMatchRange},
			
		{"firstMatch", IoRegex_firstMatch},
		{"nextMatch", IoRegex_nextMatch},
			
		{"allMatches", IoRegex_allMatches},
		{"hasMatch", IoRegex_hasMatch},
		{"version", IoRegex_version},
			
			/* --- compile options --- */
		{"caselessOn", IoRegex_caselessOn},
		{"caselessOff", IoRegex_caselessOff},
			
		{"dollarEndOnlyOn", IoRegex_dollarEndOnlyOn},
		{"dollarEndOnlyOff", IoRegex_dollarEndOnlyOff},
			
		{"dotAllOn", IoRegex_dotAllOn},
		{"dotAllOff", IoRegex_dotAllOff},
			
		{"extendedOn", IoRegex_extendedOn},
		{"extendedOff", IoRegex_extendedOff},
			
		{"multilineOn", IoRegex_multilineOn},
		{"multilineOff", IoRegex_multilineOff},
			
		{"greedyOn", IoRegex_greedyOn},
		{"greedyOff", IoRegex_greedyOff},
			
		{"UTF8On", IoRegex_UTF8On},
		{"UTF8Off", IoRegex_UTF8Off},
			
			/* --- exec options --- */
		{"matchEmptyOn", IoRegex_matchEmptyOn},
		{"matchEmptyOff", IoRegex_matchEmptyOff},
		{0x0, 0x0},
		};
		
		IoObject_addMethodTable_(self, methodTable);
	}
	
	
	return self;
}

/* ----------------------------------------------------------- */

IoRegex *IoRegex_rawClone(IoRegex *proto) 
{ 
	IoObject *self = IoObject_rawClonePrimitive(proto);
	self->data = cpalloc(proto->data, sizeof(IoRegexData));
	DATA(self)->re = 0x0; 
	DATA(self)->pe = 0x0; 
	DATA(self)->outputVector = 0x0; 
	DATA(self)->outputVectorSize = 0x0; 
	IoRegex_justUpdatePattern(self, NULL);
	return self; 
}

IoRegex *IoRegex_new(void *state)
{
	IoRegex *proto = IoState_protoWithInitFunction_(state, IoRegex_proto);
	return IOCLONE(proto);
}

void IoRegex_free(IoRegex *self) 
{ 
	if (DATA(self)->re) 
	{
		pcre_free(DATA(self)->re);
	}
	
	free(DATA(self)->outputVector);
	free(self->data); 
}

void IoRegex_mark(IoRegex *self) 
{ 
	IoObject_makeGrayIfWhite((IoObject *)DATA(self)->pattern); 
}

/* ----------------------------------------------------------- */

IoObject *IoRegex_setPattern(IoRegex *self, IoObject *locals, IoMessage *m)
{ 
	DATA(self)->pattern = IOREF(IoMessage_locals_symbolArgAt_(m, locals, 0));
	IoRegex_justUpdatePattern(self, m);
	return self;
}

void IoRegex_justUpdatePattern(IoRegex *self, IoMessage *m)
{
	char *ex;
	const char *error;
	int erroffset;
	int captures = 0;
	
	DATA(self)->lastIndex = 0;
	ex = IoSeq_asCString(DATA(self)->pattern);
	
	if (DATA(self)->re) 
	{
		pcre_free(DATA(self)->re);
	}
	
	DATA(self)->re = pcre_compile(ex, // the pattern 
		0,                        // default execOptions 
		&error,                   // for error message
		&erroffset,               // for error offset 
		NULL);                    // use default character tables
	
	if (!DATA(self)->re)
	{
		IoState_error_(IOSTATE, m, "RegularExpression compile: Unable to compile '%s' - ", ex, error);
	}
	
	// study pattern
	
	DATA(self)->pe = pcre_study(DATA(self)->re, 0, &error);
	
	if (error)
	{
		IoState_error_(IOSTATE, m, "RegularExpression study: Unable to study '%s' - %s", ex, error);
	}
	
	/* -- create output vector --- */
	pcre_fullinfo(DATA(self)->re, DATA(self)->pe, PCRE_INFO_CAPTURECOUNT, &captures);
	
	{
		int newSize = (captures + 1) * 3;
		
		if (DATA(self)->outputVectorSize != newSize)
		{
			DATA(self)->outputVectorSize = newSize;
			DATA(self)->outputVector = (int *)realloc(DATA(self)->outputVector, sizeof(int)*newSize);
			memset(DATA(self)->outputVector, 0x0, sizeof(int)*newSize);
		}
	}
}

IoObject *IoRegex_pattern(IoRegex *self, IoObject *locals, IoMessage *m)
{ 
	return DATA(self)->pattern; 
}

IoObject *IoRegex_setString(IoRegex *self, IoObject *locals, IoMessage *m)
{ 
	DATA(self)->lastIndex = 0;
	DATA(self)->string = IOREF(IoMessage_locals_symbolArgAt_(m, locals, 0));
	return self;
}

IoObject *IoRegex_string(IoRegex *self, IoObject *locals, IoMessage *m)
{ 
	return DATA(self)->string; 
}

IoObject *IoRegex_matchGroupStringAt_(IoRegex *self, int i)
{
	char *s = CSTRING(DATA(self)->string);
	int *ov = DATA(self)->outputVector;
	char *substring_start = s + ov[2*i];
	int substring_length = ov[2*i+1] - ov[2*i];
	
	if (substring_length == 0) 
	{ 
		DATA(self)->lastIndex = ov[2*i] + 1; 
	}
	else 
	{ 
		DATA(self)->lastIndex = ov[2*i] + substring_length; 
	}
	/*printf("%2d: %.*s\n", i, substring_length, substring_start);*/
	return IoState_symbolWithCString_length_(IOSTATE, substring_start, substring_length);
}

void IoRegex_findNextMatch(IoRegex *self, IoObject *locals, IoMessage *m)
{ 
	char *s = CSTRING(DATA(self)->string);
	int length = IoSeq_rawSize(DATA(self)->string);
	int rc;
	
	if (!DATA(self)->re) 
	{
		IoState_error_(IOSTATE, NULL, "RegularExpression exec: no valid pattern set\n"); 
	}
	
	rc = pcre_exec(
				DATA(self)->re,         /* the compiled pattern */
				DATA(self)->pe,         /* studied pattern */
				(const char *)s,        /* the subject string */
				length,                 /* the length of the subject */
				DATA(self)->lastIndex,          /* start at offset in the subject */
				DATA(self)->execOptions,        /* default execOptions */
				DATA(self)->outputVector,       /* output vector for substring information */
				DATA(self)->outputVectorSize);  /* number of elements in the output vector */
				
				if (rc) 
				{ 
					DATA(self)->lastIndex = *(DATA(self)->outputVector); 
				} 
				else 
				{ 
					DATA(self)->lastIndex = 0; 
				}
				
				if (rc < 0)
				{
					DATA(self)->matchGroupSize = 0;
					switch(rc)
					{
						case PCRE_ERROR_NOMATCH: 
							return;
						default: 
							/*printf("Matching error %d\n", rc);*/
							IoState_error_(IOSTATE, NULL, "RegularExpression exec: Matching error %d\n", rc); 
							break;
					}
				}
				DATA(self)->matchGroupSize = rc;
}

IoObject *IoRegex_firstMatchRange(IoRegex *self, IoObject *locals, IoMessage *m)
{
	DATA(self)->lastIndex = 0;
	return IoRegex_nextMatchRange(self, locals, m);
}

IoObject *IoRegex_nextMatchRange(IoRegex *self, IoObject *locals, IoMessage *m)
{
	IoRegex_findNextMatch(self, locals, m);
	
	if (DATA(self)->matchGroupSize)
	{
		IoList *range = IoList_new(IOSTATE);
		int *ov = DATA(self)->outputVector;
		int index  = ov[0];
		int length = ov[1] - ov[0]; 
		
		IoList_rawAppend_(range, (IoObject *)IONUMBER(index));
		IoList_rawAppend_(range, (IoObject *)IONUMBER(length));
		return range;
	}
	
	return IONIL(self);  
}

IoObject *IoRegex_firstMatch(IoRegex *self, IoObject *locals, IoMessage *m)
{
	DATA(self)->lastIndex = 0;
	return IoRegex_nextMatch(self, locals, m);
}

IoObject *IoRegex_nextMatch(IoRegex *self, IoObject *locals, IoMessage *m)
{
	IoRegex_findNextMatch(self, locals, m);
	
	if (DATA(self)->matchGroupSize == 1)
	{ 
		return IoRegex_matchGroupStringAt_(self, 0); 
	}
	else if (DATA(self)->matchGroupSize > 1)
	{
		IoList *matchGroup = IoList_new(IOSTATE);
		int i;
		
		for (i = 0; i < DATA(self)->matchGroupSize; i++)
		{ 
			IoList_rawAppend_(matchGroup, IoRegex_matchGroupStringAt_(self, i)); 
		}
		
		return matchGroup;
	}
	return IONIL(self);  
}

IoObject *IoRegex_allMatches(IoRegex *self, IoObject *locals, IoMessage *m)
{ 
	IoList *matches = IoList_new(IOSTATE);
	DATA(self)->lastIndex = 0;
	
	for (;;)
	{
		IoObject *match = IoRegex_nextMatch(self, locals, m);
		
		if (ISNIL(match)) 
		{
			break; 
		}
		
		IoList_rawAppend_(matches, match);
	} 
	
	return matches;
}

IoObject *IoRegex_hasMatch(IoRegex *self, IoObject *locals, IoMessage *m)
{
	int rc = pcre_exec(DATA(self)->re,     
				    DATA(self)->pe,  
				    (const char *)CSTRING(DATA(self)->string),     
				    (int)IoSeq_rawSize(DATA(self)->string), 
				    DATA(self)->lastIndex,
				    DATA(self)->execOptions,
				    NULL, 0);                                      
	
	if (rc == 0) 
	{ 
		return self; 
	} 
	
	return IONIL(self);
}

IoObject *IoRegex_version(IoRegex *self, IoObject *locals, IoMessage *m)
{ 
	return IOSYMBOL((char *)pcre_version()); 
}


/* --- compile execOptions ----------------------------------- */

IoObject *IoRegex_caselessOn(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->compileOptions |= PCRE_CASELESS; return self; }
IoObject *IoRegex_caselessOff(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->compileOptions &= ~PCRE_CASELESS; return self; }

IoObject *IoRegex_dollarEndOnlyOn(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->compileOptions |= PCRE_DOLLAR_ENDONLY; return self; }
IoObject *IoRegex_dollarEndOnlyOff(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->compileOptions &= ~PCRE_DOLLAR_ENDONLY; return self; }

IoObject *IoRegex_dotAllOn(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->compileOptions |= PCRE_DOTALL; return self; }
IoObject *IoRegex_dotAllOff(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->compileOptions &= ~PCRE_DOTALL; return self; }

IoObject *IoRegex_extendedOn(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->compileOptions |= PCRE_EXTENDED; return self; }
IoObject *IoRegex_extendedOff(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->compileOptions &= ~PCRE_EXTENDED; return self; }

IoObject *IoRegex_multilineOn(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->compileOptions |= PCRE_MULTILINE; return self; }
IoObject *IoRegex_multilineOff(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->compileOptions &= ~PCRE_MULTILINE; return self; }

IoObject *IoRegex_greedyOn(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->compileOptions |= PCRE_UNGREEDY; return self; }
IoObject *IoRegex_greedyOff(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->compileOptions &= ~PCRE_UNGREEDY; return self; }

IoObject *IoRegex_UTF8On(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->compileOptions |= PCRE_UTF8; return self; }
IoObject *IoRegex_UTF8Off(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->compileOptions &= ~PCRE_UTF8; return self; }

/* --- exec execOptions ----------------------------------- */

IoObject *IoRegex_matchEmptyOn(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->execOptions &= ~PCRE_NOTEMPTY; return self; }

IoObject *IoRegex_matchEmptyOff(IoRegex *self, IoObject *locals, IoMessage *m)
{ DATA(self)->execOptions |= PCRE_NOTEMPTY; return self; }

